@@ -19,6 +19,17 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* TODO:
+ *
+ * - Handle GdkEventScroll events; these come from the scrollwheel.
+ *
+ * - Rewrite scrolling mechanism.
+ *
+ * - Check if + and - work with caps lock / num lock / etc.
+ */
+
+
+
 #include <config.h>
 #include <math.h>
 #include <stdlib.h>
@@ -41,13 +52,10 @@
 #define CHECK_LIGHT 0x00aaaaaa
 #define CHECK_WHITE 0x00ffffff
 
-/* Maximum size of repaint rectangles */
+/* Maximum size of delayed repaint rectangles */
 
-#define PAINT_FAST_RECT_WIDTH 256
-#define PAINT_FAST_RECT_HEIGHT 256
-
-#define PAINT_INTERP_RECT_WIDTH 128
-#define PAINT_INTERP_RECT_HEIGHT 128
+#define PAINT_RECT_WIDTH 128
+#define PAINT_RECT_HEIGHT 128
 
 /* Scroll step increment */
 
@@ -73,9 +81,6 @@ struct _ImageViewPrivate {
 	double zoom_x_anchor;
 	double zoom_y_anchor;
 
-	/* Full screen zoom type */
-	FullScreenZoom full_screen_zoom;
-
 	/* Adjustments for scrolling */
 	GtkAdjustment *hadj;
 	GtkAdjustment *vadj;
@@ -83,11 +88,10 @@ struct _ImageViewPrivate {
 	/* Current scrolling offsets */
 	int xofs, yofs;
 
-	/* Microtile arrays for dirty region.  The first one is for immediate
-	 * drawing, the second one is for interpolated drawing.
+	/* Microtile arrays for dirty region.  This represents the dirty region
+	 * for interpolated drawing.
 	 */
-	ArtUta *uta1;
-	ArtUta *uta2;
+	ArtUta *uta;
 
 	/* Idle handler ID */
 	guint idle_id;
@@ -106,9 +110,6 @@ struct _ImageViewPrivate {
 	/* Dither type */
 	GdkRgbDither dither;
 
-	/* Scroll type */
-	ScrollType scroll;
-
 	/* Whether the image is being dragged */
 	guint dragging : 1;
 
@@ -128,9 +129,7 @@ enum {
 	PROP_INTERP_TYPE,
 	PROP_CHECK_TYPE,
 	PROP_CHECK_SIZE,
-	PROP_DITHER,
-	PROP_SCROLL,
-	PROP_FULL_SCREEN_ZOOM
+	PROP_DITHER
 };
 
 static guint image_view_signals[LAST_SIGNAL];
@@ -199,12 +198,6 @@ image_view_get_property (GObject    *object,
 	case PROP_DITHER:
 		g_value_set_int (value, priv->dither);
 		break;
-	case PROP_SCROLL:
-		g_value_set_int (value, priv->scroll);
-		break;
-	case PROP_FULL_SCREEN_ZOOM:
-		g_value_set_int (value, priv->full_screen_zoom);
-		break;
 	default:
 		g_warning ("unknown property id `%d'", property_id);
 		break;
@@ -232,12 +225,6 @@ image_view_set_property (GObject      *object,
 	case PROP_DITHER:
 		image_view_set_dither (image_view, g_value_get_int (value));
 		break;
-	case PROP_SCROLL:
-		image_view_set_scroll (image_view, g_value_get_int (value));
-		break;
-	case PROP_FULL_SCREEN_ZOOM:
-		image_view_set_full_screen_zoom (image_view, g_value_get_int (value));
-		break;
 	default:
 		g_warning ("unknown property id `%d'", property_id);
 		break;
@@ -256,7 +243,20 @@ image_view_instance_init (ImageView *view)
 	GTK_WIDGET_UNSET_FLAGS (view, GTK_NO_WINDOW);
 	GTK_WIDGET_SET_FLAGS (view, GTK_CAN_FOCUS);
 
+	priv->pixbuf = NULL;
 	priv->zoomx = priv->zoomy = 1.0;
+	priv->hadj = NULL;
+	priv->vadj = NULL;
+	priv->uta = NULL;
+
+	/* Defaults for rendering */
+	priv->interp_type = GDK_INTERP_BILINEAR;
+	priv->check_type = CHECK_TYPE_MIDTONE;
+	priv->check_size = CHECK_SIZE_LARGE;
+	priv->dither = GDK_RGB_DITHER_MAX;
+
+	/* We don't want to be double-buffered as we are SuperSmart(tm) */
+	gtk_widget_set_double_buffered (GTK_WIDGET (view), FALSE);
 }
 
 /* Frees the dirty region uta and removes the idle handler */
@@ -267,18 +267,11 @@ remove_dirty_region (ImageView *view)
 
 	priv = view->priv;
 
-	if (priv->uta1 || priv->uta2) {
+	if (priv->uta) {
 		g_assert (priv->idle_id != 0);
 
-		if (priv->uta1) {
-			art_uta_free (priv->uta1);
-			priv->uta1 = NULL;
-		}
-
-		if (priv->uta2) {
-			art_uta_free (priv->uta2);
-			priv->uta2 = NULL;
-		}
+		art_uta_free (priv->uta);
+		priv->uta = NULL;
 
 		g_source_remove (priv->idle_id);
 		priv->idle_id = 0;
@@ -428,7 +421,7 @@ static gboolean
 unity_zoom (ImageViewPrivate *priv)
 {
 	return (DOUBLE_EQUAL (priv->zoomx, 1.0) &&
-		DOUBLE_EQUAL (priv->zoomx, 1.0));
+		DOUBLE_EQUAL (priv->zoomy, 1.0));
 }
 
 /* Paints a rectangle of the dirty region */
@@ -522,7 +515,6 @@ paint_rectangle (ImageView *view, ArtIRect *rect, GdkInterpType interp_type)
 
 	/* Short-circuit the fast case to avoid a memcpy() */
 
-
 	if (unity_zoom (priv)
 	    && gdk_pixbuf_get_colorspace (priv->pixbuf) == GDK_COLORSPACE_RGB
 	    && !gdk_pixbuf_get_has_alpha (priv->pixbuf)
@@ -592,8 +584,8 @@ paint_rectangle (ImageView *view, ArtIRect *rect, GdkInterpType interp_type)
 		break;
 
 	default:
-		check_1 = CHECK_DARK;
-		check_2 = CHECK_LIGHT;
+		g_assert_not_reached ();
+		return;
 	}
 
 	switch (priv->check_size) {
@@ -610,8 +602,8 @@ paint_rectangle (ImageView *view, ArtIRect *rect, GdkInterpType interp_type)
 		break;
 
 	default:
-		check_size = CHECK_LARGE;
-		break;
+		g_assert_not_reached ();
+		return;
 	}
 
 	/* Draw! */
@@ -671,51 +663,17 @@ paint_iteration_idle (gpointer data)
 	view = IMAGE_VIEW (data);
 	priv = view->priv;
 
-	g_assert (priv->uta1 != NULL || priv->uta2 != NULL);
+	g_assert (priv->uta != NULL);
 
-	if (priv->uta1) {
-		/* Paint the no-interpolation cases as fast as possible */
+	pull_rectangle (priv->uta, &rect, PAINT_RECT_WIDTH, PAINT_RECT_HEIGHT);
 
-		if (priv->scroll == SCROLL_TWO_PASS || priv->interp_type == GDK_INTERP_NEAREST)
-			while (1) {
-				pull_rectangle (priv->uta1, &rect,
-						PAINT_FAST_RECT_WIDTH, PAINT_FAST_RECT_HEIGHT);
+	if (art_irect_empty (&rect)) {
+		art_uta_free (priv->uta);
+		priv->uta = NULL;
+	} else
+		paint_rectangle (view, &rect, priv->interp_type);
 
-				if (art_irect_empty (&rect)) {
-					art_uta_free (priv->uta1);
-					priv->uta1 = NULL;
-					break;
-				}
-
-				paint_rectangle (view, &rect, GDK_INTERP_NEAREST);
-
-				if (priv->scroll == SCROLL_TWO_PASS)
-					priv->uta2 = uta_add_rect (priv->uta2,
-								   rect.x0, rect.y0,
-								   rect.x1, rect.y1);
-			}
-		else {
-			pull_rectangle (priv->uta1, &rect,
-					PAINT_INTERP_RECT_WIDTH, PAINT_INTERP_RECT_HEIGHT);
-
-			if (art_irect_empty (&rect)) {
-				art_uta_free (priv->uta1);
-				priv->uta1 = NULL;
-			} else
-				paint_rectangle (view, &rect, priv->interp_type);
-		}
-	} else if (priv->uta2) {
-		pull_rectangle (priv->uta2, &rect,
-				PAINT_INTERP_RECT_WIDTH, PAINT_INTERP_RECT_HEIGHT);
-
-		if (art_irect_empty (&rect)) {
-			art_uta_free (priv->uta2);
-			priv->uta2 = NULL;
-		} else
-			paint_rectangle (view, &rect, priv->interp_type);
-	}
-
-	if (!priv->uta1 && !priv->uta2) {
+	if (!priv->uta) {
 		priv->idle_id = 0;
 		return FALSE;
 	}
@@ -723,9 +681,12 @@ paint_iteration_idle (gpointer data)
 	return TRUE;
 }
 
-/* Queues a repaint of the specified area in window coordinates */
+/* Paints the requested area in non-interpolated mode.  Then, if we are
+ * configured to use interpolation, we queue an idle handler to redraw the area
+ * with interpolation.  The area is in window coordinates.
+ */
 static void
-request_paint_area (ImageView *view, GdkRectangle *area, gboolean asynch)
+request_paint_area (ImageView *view, GdkRectangle *area)
 {
 	ImageViewPrivate *priv;
 	ArtIRect r;
@@ -743,49 +704,46 @@ request_paint_area (ImageView *view, GdkRectangle *area, gboolean asynch)
 	if (r.x0 >= r.x1 || r.y0 >= r.y1)
 		return;
 
-	/* Unless told not to, do nearest neighbor or 1:1 zoom synchronously for
-         * speed.
-	 */
+	/* Do nearest neighbor or 1:1 zoom synchronously for speed.  */
 
-	if (!asynch && (priv->interp_type == GDK_INTERP_NEAREST || unity_zoom (priv))) {
+	if (priv->interp_type == GDK_INTERP_NEAREST || unity_zoom (priv)) {
 		paint_rectangle (view, &r, priv->interp_type);
 		return;
 	}
 
-	/* All other interpolation types are delayed */
+	/* All other interpolation types are delayed.  */
 
-	if (priv->uta1 || priv->uta2)
+	if (priv->uta)
 		g_assert (priv->idle_id != 0);
 	else {
 		g_assert (priv->idle_id == 0);
 		priv->idle_id = g_idle_add (paint_iteration_idle, view);
 	}
 
-	if (asynch || priv->scroll != SCROLL_TWO_PASS)
-		priv->uta1 = uta_add_rect (priv->uta1, r.x0, r.y0, r.x1, r.y1);
-	else {
-		paint_rectangle (view, &r, GDK_INTERP_NEAREST);
-		priv->uta2 = uta_add_rect (priv->uta2, r.x0, r.y0, r.x1, r.y1);
-	}
+	paint_rectangle (view, &r, GDK_INTERP_NEAREST);
+	priv->uta = uta_add_rect (priv->uta, r.x0, r.y0, r.x1, r.y1);
 }
 
-/* Scrolls the view to the specified offsets.  Does not perform range checking!  */
+/* Scrolls the view to the specified offsets.  */
 static void
-scroll_to (ImageView *view, int x, int y)
+scroll_to (ImageView *view, int x, int y, gboolean change_adjustments)
 {
 	ImageViewPrivate *priv;
 	int xofs, yofs;
 	GdkWindow *window;
-	GdkGC *gc;
 	int width, height;
 	int src_x, src_y;
 	int dest_x, dest_y;
 	int twidth, theight;
-	GdkEvent *event;
 
 	priv = view->priv;
 
-	/* Compute offsets and check bounds */
+	/* Check bounds */
+
+	x = CLAMP (x, 0, priv->hadj->upper - priv->hadj->page_size);
+	y = CLAMP (y, 0, priv->vadj->upper - priv->vadj->page_size);
+
+	/* Compute offsets */
 
 	xofs = x - priv->xofs;
 	yofs = y - priv->yofs;
@@ -797,7 +755,7 @@ scroll_to (ImageView *view, int x, int y)
 	priv->yofs = y;
 
 	if (!GTK_WIDGET_DRAWABLE (view))
-		return;
+		goto out;
 
 	width = GTK_WIDGET (view)->allocation.width;
 	height = GTK_WIDGET (view)->allocation.height;
@@ -810,8 +768,8 @@ scroll_to (ImageView *view, int x, int y)
 		area.width = width;
 		area.height = height;
 
-		request_paint_area (view, &area, FALSE);
-		return;
+		request_paint_area (view, &area);
+		goto out;
 	}
 
 	window = GTK_WIDGET (view)->window;
@@ -821,80 +779,68 @@ scroll_to (ImageView *view, int x, int y)
 	twidth = (width + ART_UTILE_SIZE - 1) >> ART_UTILE_SHIFT;
 	theight = (height + ART_UTILE_SIZE - 1) >> ART_UTILE_SHIFT;
 
-	if (priv->uta1 || priv->uta2)
+	if (priv->uta)
 		g_assert (priv->idle_id != 0);
 	else
 		priv->idle_id = g_idle_add (paint_iteration_idle, view);
 
-	priv->uta1 = uta_ensure_size (priv->uta1, 0, 0, twidth, theight);
+	priv->uta = uta_ensure_size (priv->uta, 0, 0, twidth, theight);
 
-	/* Copy the uta area */
+	/* Copy the uta area.  Our synchronous handling of expose events, below,
+	 * will queue the new scrolled-in areas.
+	 */
 
 	src_x = xofs < 0 ? 0 : xofs;
 	src_y = yofs < 0 ? 0 : yofs;
 	dest_x = xofs < 0 ? -xofs : 0;
 	dest_y = yofs < 0 ? -yofs : 0;
 
-	uta_copy_area (priv->uta1,
+	uta_copy_area (priv->uta,
 		       src_x, src_y,
 		       dest_x, dest_y,
 		       width - abs (xofs), height - abs (yofs));
 
-	if (priv->scroll == SCROLL_TWO_PASS && priv->uta2) {
-		priv->uta2 = uta_ensure_size (priv->uta2, 0, 0, twidth, theight);
+	/* Scroll the window area and process exposure synchronously. */
 
-		uta_copy_area (priv->uta2,
-			       src_x, src_y,
-			       dest_x, dest_y,
-			       width - abs (xofs), height - abs (yofs));
-	}
+	gdk_window_scroll (window, -xofs, -yofs);
+	gdk_window_process_updates (window, TRUE);
 
-	/* Copy the window area */
+ out:
+	if (!change_adjustments)
+		return;
 
-	gc = gdk_gc_new (window);
-	gdk_gc_set_exposures (gc, TRUE);
+	g_signal_handlers_block_matched (
+		priv->hadj, G_SIGNAL_MATCH_DATA,
+		0, 0, NULL, NULL, view);
+	g_signal_handlers_block_matched (
+		priv->vadj, G_SIGNAL_MATCH_DATA,
+		0, 0, NULL, NULL, view);
 
-	gdk_draw_drawable (window, gc, window, src_x, src_y,
-			   dest_x, dest_y, 
-			   width - abs (xofs),
-			   height - abs (yofs));
+	priv->hadj->value = x;
+	priv->vadj->value = y;
 
-	g_object_unref (gc);
+	g_signal_emit_by_name (priv->hadj, "value_changed");
+	g_signal_emit_by_name (priv->vadj, "value_changed");
 
-	/* Add the scrolled-in region */
+	g_signal_handlers_unblock_matched (
+		priv->hadj, G_SIGNAL_MATCH_DATA,
+		0, 0, NULL, NULL, view);
+	g_signal_handlers_unblock_matched (
+		priv->vadj, G_SIGNAL_MATCH_DATA,
+		0, 0, NULL, NULL, view);
+}
 
-	if (xofs) {
-		GdkRectangle r;
+/* Scrolls the image view by the specified offsets.  Notifies the adjustments
+ * about their new values.
+ */
+static void
+scroll_by (ImageView *view, int xofs, int yofs)
+{
+	ImageViewPrivate *priv;
 
-		r.x = xofs < 0 ? 0 : width - xofs;
-		r.y = 0;
-		r.width = abs (xofs);
-		r.height = height;
+	priv = view->priv;
 
-		request_paint_area (view, &r, FALSE);
-	}
-
-	if (yofs) {
-		GdkRectangle r;
-
-		r.x = 0;
-		r.y = yofs < 0 ? 0 : height - yofs;
-		r.width = width;
-		r.height = abs (yofs);
-
-		request_paint_area (view, &r, FALSE);
-	}
-
-	/* Process graphics exposures */
-
-	while ((event = gdk_event_get_graphics_expose (window)) != NULL) {
-		gtk_widget_event (GTK_WIDGET (view), event);
-		if (event->expose.count == 0) {
-			gdk_event_free (event);
-			break;
-		}
-		gdk_event_free (event);
-	}
+	scroll_to (view, priv->xofs + xofs, priv->yofs + yofs, TRUE);
 }
 
 
@@ -938,6 +884,10 @@ image_view_realize (GtkWidget *widget)
 	attr.event_mask = (gtk_widget_get_events (widget)
 			   | GDK_EXPOSURE_MASK
 			   | GDK_BUTTON_PRESS_MASK
+			   | GDK_BUTTON_RELEASE_MASK
+			   | GDK_POINTER_MOTION_MASK
+			   | GDK_POINTER_MOTION_HINT_MASK
+			   | GDK_SCROLL_MASK
 			   | GDK_KEY_PRESS_MASK);
 
 	attr_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
@@ -1022,7 +972,8 @@ compute_center_zoom_offsets (ImageView *view,
 	priv = view->priv;
 	g_assert (priv->need_zoom_change);
 
-	compute_scaled_size (view, priv->old_zoomx, priv->old_zoomy, &old_scaled_width, &old_scaled_height);
+	compute_scaled_size (view, priv->old_zoomx, priv->old_zoomy,
+			     &old_scaled_width, &old_scaled_height);
 
 	if (old_scaled_width < old_width)
 		view_cx = (priv->zoom_x_anchor * old_scaled_width) / priv->old_zoomx;
@@ -1034,7 +985,8 @@ compute_center_zoom_offsets (ImageView *view,
 	else
 		view_cy = (priv->yofs + priv->zoom_y_anchor * old_height) / priv->old_zoomy;
 
-	compute_scaled_size (view, priv->zoomx, priv->zoomy, &new_scaled_width, &new_scaled_height);
+	compute_scaled_size (view, priv->zoomx, priv->zoomy,
+			     &new_scaled_width, &new_scaled_height);
 
 	if (new_scaled_width < new_width)
 		*xofs = 0;
@@ -1146,14 +1098,13 @@ image_view_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 	}
 }
 
-/* Button press handler for the image view */
-static gint
-image_view_button_press (GtkWidget *widget, GdkEventButton *event)
+/* Button press event handler for the image view */
+static gboolean
+image_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 {
 	ImageView *view;
 	ImageViewPrivate *priv;
 	GdkCursor *cursor;
-	int retval;
 
 	view = IMAGE_VIEW (widget);
 	priv = view->priv;
@@ -1167,19 +1118,9 @@ image_view_button_press (GtkWidget *widget, GdkEventButton *event)
 	switch (event->button) {
 	case 1:
 		cursor = cursor_get (widget, CURSOR_HAND_CLOSED);
-		retval = gdk_pointer_grab (widget->window,
-					   FALSE,
-					   (GDK_POINTER_MOTION_MASK
-					    | GDK_POINTER_MOTION_HINT_MASK
-					    | GDK_BUTTON_RELEASE_MASK),
-					   NULL,
-					   cursor,
-					   event->time);
+		gdk_window_set_cursor (widget->window, cursor);
 		gdk_cursor_unref (cursor);
 
-		if (retval != 0)
-			return FALSE;
-
 		priv->dragging = TRUE;
 		priv->drag_anchor_x = event->x;
 		priv->drag_anchor_y = event->y;
@@ -1218,38 +1159,19 @@ drag_to (ImageView *view, int x, int y)
 	dx = priv->drag_anchor_x - x;
 	dy = priv->drag_anchor_y - y;
 
-	x = CLAMP (priv->drag_ofs_x + dx, 0, priv->hadj->upper - priv->hadj->page_size);
-	y = CLAMP (priv->drag_ofs_y + dy, 0, priv->vadj->upper - priv->vadj->page_size);
+	x = priv->drag_ofs_x + dx;
+	y = priv->drag_ofs_y + dy;
 
-	scroll_to (view, x, y);
-	
-	g_signal_handlers_block_matched (
-		priv->hadj, G_SIGNAL_MATCH_DATA,
-		0, 0, NULL, NULL, view);
-	g_signal_handlers_block_matched (
-		priv->vadj, G_SIGNAL_MATCH_DATA,
-		0, 0, NULL, NULL, view);
-
-	priv->hadj->value = x;
-	priv->vadj->value = y;
-
-	g_signal_emit_by_name (priv->hadj, "value_changed");
-	g_signal_emit_by_name (priv->vadj, "value_changed");
-
-	g_signal_handlers_unblock_matched (
-		priv->hadj, G_SIGNAL_MATCH_DATA,
-		0, 0, NULL, NULL, view);
-	g_signal_handlers_unblock_matched (
-		priv->vadj, G_SIGNAL_MATCH_DATA,
-		0, 0, NULL, NULL, view);
+	scroll_to (view, x, y, TRUE);
 }
 
-/* Button release handler for the image view */
-static gint
-image_view_button_release (GtkWidget *widget, GdkEventButton *event)
+/* Button release event handler for the image view */
+static gboolean
+image_view_button_release_event (GtkWidget *widget, GdkEventButton *event)
 {
 	ImageView *view;
 	ImageViewPrivate *priv;
+	GdkCursor *cursor;
 
 	view = IMAGE_VIEW (widget);
 	priv = view->priv;
@@ -1259,14 +1181,77 @@ image_view_button_release (GtkWidget *widget, GdkEventButton *event)
 
 	drag_to (view, event->x, event->y);
 	priv->dragging = FALSE;
-	gdk_pointer_ungrab (event->time);
+
+	cursor = cursor_get (widget, CURSOR_HAND_OPEN);
+	gdk_window_set_cursor (widget->window, cursor);
+	gdk_cursor_unref (cursor);
 
 	return TRUE;
 }
 
-/* Motion handler for the image view */
-static gint
-image_view_motion (GtkWidget *widget, GdkEventMotion *event)
+/* Scroll event handler for the image view.  We zoom with an event without
+ * modifiers rather than scroll; we use the Shift modifier to scroll.
+ * Rationale: images are not primarily vertical, and in EOG you scan scroll by
+ * dragging the image with button 1 anyways.
+ */
+static gboolean
+image_view_scroll_event (GtkWidget *widget, GdkEventScroll *event)
+{
+	ImageView *view;
+	ImageViewPrivate *priv;
+	double zoom_factor;
+	int xofs, yofs;
+
+	view = IMAGE_VIEW (widget);
+	priv = view->priv;
+
+	/* Compute zoom factor and scrolling offsets; we'll only use either of them */
+
+	xofs = priv->hadj->page_increment / 2; /* same as in gtkscrolledwindow.c */
+	yofs = priv->vadj->page_increment / 2;
+
+	switch (event->direction) {
+	case GDK_SCROLL_UP:
+		zoom_factor = IMAGE_VIEW_ZOOM_MULTIPLIER;
+		xofs = 0;
+		yofs = -yofs;
+		break;
+
+	case GDK_SCROLL_LEFT:
+		zoom_factor = 1.0 / IMAGE_VIEW_ZOOM_MULTIPLIER;
+		xofs = -xofs;
+		yofs = 0;
+		break;
+
+	case GDK_SCROLL_DOWN:
+		zoom_factor = 1.0 / IMAGE_VIEW_ZOOM_MULTIPLIER;
+		xofs = 0;
+		yofs = yofs;
+		break;
+
+	case GDK_SCROLL_RIGHT:
+		zoom_factor = IMAGE_VIEW_ZOOM_MULTIPLIER;
+		xofs = xofs;
+		yofs = 0;
+		break;
+
+	default:
+		g_assert_not_reached ();
+		return FALSE;
+	}
+
+	if ((event->state & GDK_SHIFT_MASK) == 0) {
+		set_zoom_anchor (view, event->x, event->y);
+		image_view_set_zoom (view, priv->zoomx * zoom_factor, priv->zoomy * zoom_factor);
+	} else
+		scroll_by (view, xofs, yofs);
+
+	return TRUE;
+}
+
+/* Motion event handler for the image view */
+static gboolean
+image_view_motion_event (GtkWidget *widget, GdkEventMotion *event)
 {
 	ImageView *view;
 	ImageViewPrivate *priv;
@@ -1290,11 +1275,18 @@ image_view_motion (GtkWidget *widget, GdkEventMotion *event)
 	return TRUE;
 }
 
-/* Expose handler for the image view */
-static gint
-image_view_expose (GtkWidget *widget, GdkEventExpose *event)
+/* Expose event handler for the image view.  First we process the whole dirty
+ * region by drawing a non-interpolated version, which is "instantaneous", and
+ * we do this synchronously.  Then, if we are set to use interpolation, we queue
+ * an idle handler to handle interpolated drawing there.
+ */
+static gboolean
+image_view_expose_event (GtkWidget *widget, GdkEventExpose *event)
 {
 	ImageView *view;
+	GdkRectangle *rects;
+	gint n_rects;
+	int i;
 
 	g_return_val_if_fail (widget != NULL, FALSE);
 	g_return_val_if_fail (IS_IMAGE_VIEW (widget), FALSE);
@@ -1302,13 +1294,19 @@ image_view_expose (GtkWidget *widget, GdkEventExpose *event)
 
 	view = IMAGE_VIEW (widget);
 
-	request_paint_area (view, &event->area, FALSE);
+	gdk_region_get_rectangles (event->region, &rects, &n_rects);
+
+	for (i = 0; i < n_rects; i++)
+		request_paint_area (view, rects + i);
+
+	g_free (rects);
+
 	return TRUE;
 }
 
-/* Key press handler for the image view */
-static gint
-image_view_key_press (GtkWidget *widget, GdkEventKey *event)
+/* Key press event handler for the image view */
+static gboolean
+image_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 {
 	ImageView *view;
 	ImageViewPrivate *priv;
@@ -1395,34 +1393,8 @@ image_view_key_press (GtkWidget *widget, GdkEventKey *event)
 		image_view_set_zoom (view, zoomx, zoomy);
 	}
 
-	if (do_scroll) {
-		int x, y;
-
-		x = CLAMP (priv->xofs + xofs, 0, priv->hadj->upper - priv->hadj->page_size);
-		y = CLAMP (priv->yofs + yofs, 0, priv->vadj->upper - priv->vadj->page_size);
-
-		scroll_to (view, x, y);
-
-		g_signal_handlers_block_matched (
-			priv->hadj, G_SIGNAL_MATCH_DATA,
-			0, 0, NULL, NULL, view);
-		g_signal_handlers_block_matched (
-			priv->vadj, G_SIGNAL_MATCH_DATA,
-			0, 0, NULL, NULL, view);
-
-		priv->hadj->value = x;
-		priv->vadj->value = y;
-
-		g_signal_emit_by_name (priv->hadj, "value_changed");
-		g_signal_emit_by_name (priv->vadj, "value_changed");
-
-		g_signal_handlers_unblock_matched (
-			priv->hadj, G_SIGNAL_MATCH_DATA,
-			0, 0, NULL, NULL, view);
-		g_signal_handlers_unblock_matched (
-			priv->vadj, G_SIGNAL_MATCH_DATA,
-			0, 0, NULL, NULL, view);
-	}
+	if (do_scroll)
+		scroll_by (view, xofs, yofs);
 
 	return TRUE;
 }
@@ -1437,7 +1409,7 @@ adjustment_changed_cb (GtkAdjustment *adj, gpointer data)
 	view = IMAGE_VIEW (data);
 	priv = view->priv;
 
-	scroll_to (view, priv->hadj->value, priv->vadj->value);
+	scroll_to (view, priv->hadj->value, priv->vadj->value, FALSE);
 }
 
 /* Set_scroll_adjustments handler for the image view */
@@ -1576,7 +1548,8 @@ image_view_set_pixbuf (ImageView *view, GdkPixbuf *pixbuf)
 
 	view->priv->pixbuf = pixbuf;
 
-	/* FIXME: adjust zoom / image offsets; maybe just offsets here */
+	remove_dirty_region (view);
+	scroll_to (view, 0, 0, TRUE);
 
 	gtk_widget_queue_draw (GTK_WIDGET (view));
 }
@@ -1630,21 +1603,26 @@ image_view_set_zoom (ImageView *view, double zoomx, double zoomy)
 /**
  * image_view_get_zoom:
  * @view: An image view.
+ * @zoomx: If non-NULL, the horizontal zoom factor is returned here.
+ * @zoomy: If non-NULL, the vertical zoom factor is returned here.
  *
  * Queries the zoom factor of an image view.
- *
- * Return value: Current zoom factor.
  **/
-double
-image_view_get_zoom (ImageView *view)
+void
+image_view_get_zoom (ImageView *view, double *zoomx, double *zoomy)
 {
 	ImageViewPrivate *priv;
 
-	g_return_val_if_fail (view != NULL, -1.0);
-	g_return_val_if_fail (IS_IMAGE_VIEW (view), -1.0);
+	g_return_if_fail (view != NULL);
+	g_return_if_fail (IS_IMAGE_VIEW (view));
 
 	priv = view->priv;
-	return (priv->zoomx + priv->zoomy) / 2;
+
+	if (zoomx)
+		*zoomx = priv->zoomx;
+
+	if (zoomy)
+		*zoomy = priv->zoomy;
 }
 
 /**
@@ -1826,96 +1804,6 @@ image_view_get_dither (ImageView *view)
 	return priv->dither;
 }
 
-/**
- * image_view_set_scroll:
- * @view: An image view.
- * @scroll: Scrolling type.
- *
- * Sets the scrolling type on an image view.
- **/
-void
-image_view_set_scroll (ImageView *view, ScrollType scroll)
-{
-	ImageViewPrivate *priv;
-
-	g_return_if_fail (view != NULL);
-	g_return_if_fail (IS_IMAGE_VIEW (view));
-
-	priv = view->priv;
-
-	if (priv->scroll == scroll)
-		return;
-
-	priv->scroll = scroll;
-
-	gtk_widget_queue_draw (GTK_WIDGET (view));
-}
-
-/**
- * image_view_get_scroll:
- * @view: An image view.
- *
- * Queries the scrolling type of an image view.
- *
- * Return value: Scrolling type.
- **/
-ScrollType
-image_view_get_scroll (ImageView *view)
-{
-	ImageViewPrivate *priv;
-
-	g_return_val_if_fail (view != NULL, SCROLL_NORMAL);
-	g_return_val_if_fail (IS_IMAGE_VIEW (view), SCROLL_NORMAL);
-
-	priv = view->priv;
-	return priv->scroll;
-}
-
-/**
- * image_view_set_full_screen_zoom:
- * @view: An image view.
- * @full_screen_zoom: Full screen zooming type.
- *
- * Sets the full screen zooming type on an image view.
- **/
-void
-image_view_set_full_screen_zoom (ImageView *view, FullScreenZoom full_screen_zoom)
-{
-	ImageViewPrivate *priv;
-
-	g_return_if_fail (view != NULL);
-	g_return_if_fail (IS_IMAGE_VIEW (view));
-
-	priv = view->priv;
-
-	if (priv->full_screen_zoom == full_screen_zoom)
-		return;
-
-	priv->full_screen_zoom = full_screen_zoom;
-
-	gtk_widget_queue_draw (GTK_WIDGET (view));
-}
-
-/**
- * image_view_get_full_screen_zoom:
- * @view: An image view.
- *
- * Queries the full_screen_zooming type of an image view.
- *
- * Return value: full screen zooming type.
- **/
-FullScreenZoom
-image_view_get_full_screen_zoom (ImageView *view)
-{
-	ImageViewPrivate *priv;
-
-	g_return_val_if_fail (view != NULL, FULL_SCREEN_ZOOM_1);
-	g_return_val_if_fail (IS_IMAGE_VIEW (view), FULL_SCREEN_ZOOM_1);
-
-	priv = view->priv;
-	return priv->full_screen_zoom;
-}
-
 /**
  * image_view_get_scaled_size
  * @view: An image view.
@@ -1986,23 +1874,8 @@ image_view_class_init (ImageViewClass *class)
 				  _("dither"),
 				  _("dither type"),
 				  0, G_MAXINT, 0, 0));
-	g_object_class_install_property (
-		gobject_class,
-		PROP_SCROLL,
-		g_param_spec_int ("scroll",
-				  _("scroll"),
-				  _("scroll type"),
-				  0, G_MAXINT, 0, 0));
-	g_object_class_install_property (
-		gobject_class,
-		PROP_FULL_SCREEN_ZOOM,
-		g_param_spec_int ("full_screen_zoom",
-				  _("enable full screen zoom"),
-				  _("whether we should allow full screen zoom"),
-				  0, G_MAXINT, 0, 0));
-
   	image_view_signals[ZOOM_FIT] =
- 		g_signal_new ("zoom_fit", 
+ 		g_signal_new ("zoom_fit",
 			      G_TYPE_FROM_CLASS (object_class),
 			      G_SIGNAL_RUN_FIRST,
 			      G_STRUCT_OFFSET (ImageViewClass, zoom_fit),
@@ -2021,10 +1894,10 @@ image_view_class_init (ImageViewClass *class)
 			      g_cclosure_marshal_VOID__VOID,
 			      G_TYPE_NONE,
 			      0);
-  
+
 	gobject_class->dispose = image_view_dispose;
   	gobject_class->finalize = image_view_finalize;
-  
+
   	class->set_scroll_adjustments = image_view_set_scroll_adjustments;
   	widget_class->set_scroll_adjustments_signal =
  		g_signal_new ("set_scroll_adjustments",
@@ -2044,11 +1917,12 @@ image_view_class_init (ImageViewClass *class)
 	widget_class->unrealize = image_view_unrealize;
 	widget_class->size_request = image_view_size_request;
 	widget_class->size_allocate = image_view_size_allocate;
-	widget_class->button_press_event = image_view_button_press;
-	widget_class->button_release_event = image_view_button_release;
-	widget_class->motion_notify_event = image_view_motion;
-	widget_class->expose_event = image_view_expose;
-	widget_class->key_press_event = image_view_key_press;
+	widget_class->button_press_event = image_view_button_press_event;
+	widget_class->button_release_event = image_view_button_release_event;
+	widget_class->scroll_event = image_view_scroll_event;
+	widget_class->motion_notify_event = image_view_motion_event;
+	widget_class->expose_event = image_view_expose_event;
+	widget_class->key_press_event = image_view_key_press_event;
 	widget_class->focus_in_event  = image_view_focus_in_event;
 	widget_class->focus_out_event = image_view_focus_out_event;
 }
